"""
Common utilities and helpers.

"""

__all__ = ['compute_histogram', 'preprocess_histogram', 'compute_edge_weight', 'HiwenetWarning']

import numpy as np

def compute_histogram(values, edges, use_orig_distr=False):
    """Computes histogram (density) for a given vector of values."""

    if use_orig_distr:
        return values

    # ignoring invalid values: Inf and Nan
    values = check_array(values).compressed()

    hist, bin_edges = np.histogram(values, bins=edges, density=True)
    hist = preprocess_histogram(hist, values, edges)

    return hist


def preprocess_histogram(hist, values, edges):
    """Handles edge-cases and extremely-skewed histograms"""

    # working with extremely skewed histograms
    if np.count_nonzero(hist) == 0:
        # all of them above upper bound
        if np.all(values >= edges[-1]):
            hist[-1] = 1
        # all of them below lower bound
        elif np.all(values <= edges[0]):
            hist[0] = 1

    return hist


def compute_edge_weight(hist_one, hist_two, weight_func):
    """
    Computes the edge weight between the two histograms.

    Parameters
    ----------
    hist_one : sequence
        First histogram

    hist_two : sequence
        Second histogram

    weight_func : callable
        Identifying the type of distance (or metric) to compute between the pair of histograms.
        Must be one of the metrics implemented in medpy.metric.histogram, or another valid callable.

    Returns
    -------
    edge_value : float
        Distance or metric between the two histograms
    """

    edge_value = weight_func(hist_one, hist_two)

    return edge_value


class HiwenetWarning(Warning):
    """Exception to indicate the detection of non-fatal use of hiwenet."""
    pass


def check_array(array):
    "Converts to flattened numpy arrays and ensures its not empty."

    if len(array) < 1:
        raise ValueError('Input array is empty! Must have atleast 1 element.')

    return np.ma.masked_invalid(array).flatten()


"""
Unit and Hypothesis Tests for histogram metrics

"""

import numpy as np
from hypothesis import given, strategies, assume, Verbosity, note, event
from hypothesis import settings as hyp_settings

from medpy.metric import histogram

metric_list = ['manhattan', 'minowski', 'euclidean', 'noelle_2', 'noelle_4', 'noelle_5']
metric_list_to_doublecheck = ['cosine_1']

unknown_property = ['histogram_intersection']
still_under_dev  = ['quadratic_forms']
similarity_funcs = ['correlate', 'cosine', 'cosine_2', 'cosine_alt', 'fidelity_based']
semi_metric_list = ['kullback_leibler', 'jensen_shannon', 'chi_square', 'chebyshev', 'chebyshev_neg',
                    'histogram_intersection_1', 'relative_deviation', 'relative_bin_deviation',
                    'noelle_1', 'noelle_3', 'correlate_1']

default_feature_dim = 1000
default_num_bins = 20

range_feature_dim = [10, 10000]
range_num_bins = [5, 200]


def within_tolerance(x, y):
    "Function to indicate acceptable level of tolerance in numerical differences"

    # as the np.allcose function is not symmetric,
    # this ensurers second arg is larger
    if x < y:
        smaller = x
        larger = y
    else:
        smaller = y
        larger = x

    # atol=np.finfo(float).eps is super strict, choosing 1e-3
    # rtol=1e-2 approx. matches two decimal points
    return bool(np.allclose(smaller, larger, rtol=1e-2, atol=1e-3))


def make_random_histogram(length=default_feature_dim, num_bins=default_num_bins):
    "Returns a sequence of histogram density values that sum to 1.0"

    hist, bin_edges = np.histogram(np.random.random(length),
                                   bins=num_bins, density=True)

    # to ensure they sum to 1.0
    hist = hist / sum(hist)

    if len(hist) < 2:
        raise ValueError('Invalid histogram')

    return hist


# Increasing the number of examples to try
@hyp_settings(max_examples=1000, min_satisfying_examples=100)  # , verbosity=Verbosity.verbose)
@given(strategies.sampled_from(metric_list),
       strategies.integers(range_feature_dim[0], range_feature_dim[1]),
       strategies.integers(range_num_bins[0], range_num_bins[1]))
def test_math_properties_metric(method_str, feat_dim, num_bins):
    """Trying to test the four properties on the same set of histograms"""

    # checking for bad examples
    assume(not num_bins > feat_dim)

    h1 = make_random_histogram(feat_dim, num_bins)
    h2 = make_random_histogram(feat_dim, num_bins)
    h3 = make_random_histogram(feat_dim, num_bins)

    method = getattr(histogram, method_str)

    check_indiscernibility(method, h1)
    check_symmetry(method, h1, h2)
    check_nonnegativity(method, h1, h2)
    check_triangle_inequality(method, h1, h2, h3)


def check_indiscernibility(method, hist):
    """a must be unique, and a is identical to a if and only if dist(a,a)=0"""

    assert within_tolerance(method(hist, hist), 0.0)


def check_symmetry(method, h1, h2):
    """symmetry test"""

    d12 = method(h1, h2)
    d21 = method(h2, h1)

    assert within_tolerance(d12, d21)


def check_nonnegativity(method, h1, h2):
    """distance between two samples must be >= 0.0"""

    assert method(h1, h2) >= 0.0


def check_triangle_inequality(method, h1, h2, h3):
    """ Classic test for a metric: dist(a,b) < dist(a,b) + dist(a,c)"""

    d12 = method(h1, h2)
    d23 = method(h2, h3)
    d13 = method(h1, h3)

    d13_plus_d23 = np.round(d13 + d23, decimals=10)
    d12_ = np.round(d12, decimals=10)
    assert d12_ <= d13_plus_d23



"""
Module implementing additional metrics for edge weights.

"""

from sys import version_info
if version_info.major==2 and version_info.minor==7:
    from .utils import check_array
elif version_info.major > 2:
    from hiwenet.utils import check_array
else:
    raise NotImplementedError('hiwenet supports only 2.7 or 3+. '
                              'Upgrade to Python 3+ is recommended.')

__all__ = ['diff_medians', 'diff_medians_abs',
           'diff_means', 'diff_means_abs']

import numpy as np


def diff_medians(array_one, array_two):
    """
    Computes the difference in medians between two arrays of values.

    Given arrays will be flattened (to 1D array) regardless of dimension,
        and any non-finite/NaN values will be ignored.

    Parameters
    ----------
    array_one, array_two : iterable
        Two arrays of values, possibly of different length.

    Returns
    -------
    diff_medians : float
        scalar measuring the difference in medians, ignoring NaNs/non-finite values.

    Raises
    ------
    ValueError
        If one or more of the arrays are empty.

    """

    array_one = check_array(array_one)
    array_two = check_array(array_two)
    diff_medians = np.ma.median(array_one) - np.ma.median(array_two)

    return diff_medians


def diff_medians_abs(array_one, array_two):
    """
    Computes the absolute (symmetric) difference in medians between two arrays of values.

    Given arrays will be flattened (to 1D array) regardless of dimension,
        and any non-finite/NaN values will be ignored.

    Parameters
    ----------
    array_one, array_two : iterable
        Two arrays of values, possibly of different length.

    Returns
    -------
    diff_medians : float
        scalar measuring the difference in medians, ignoring NaNs/non-finite values.

    Raises
    ------
    ValueError
        If one or more of the arrays are empty.

    """

    abs_diff_medians = np.abs(diff_medians(array_one, array_two))

    return abs_diff_medians


def diff_means(array_one, array_two):
    """
    Computes the difference in means between two arrays of values.

    Given arrays will be flattened (to 1D array) regardless of dimension,
        and any non-finite/NaN values will be ignored.

    Parameters
    ----------
    array_one, array_two : iterable
        Two arrays of values, possibly of different length.

    Returns
    -------
    diff_medians : float
        scalar measuring the difference in medians, ignoring NaNs/non-finite values.

    Raises
    ------
    ValueError
        If one or more of the arrays are empty.

    """

    array_one = check_array(array_one)
    array_two = check_array(array_two)
    diff_means = np.ma.mean(array_one) - np.ma.mean(array_two)

    return diff_means


def diff_means_abs(array_one, array_two):
    """
    Computes the absolute (symmetric) difference in means between two arrays of values.

    Given arrays will be flattened (to 1D array) regardless of dimension,
        and any non-finite/NaN values will be ignored.

    Parameters
    ----------
    array_one, array_two : iterable
        Two arrays of values, possibly of different length.

    Returns
    -------
    diff_medians : float
        scalar measuring the difference in medians, ignoring NaNs/non-finite values.

    Raises
    ------
    ValueError
        If one or more of the arrays are empty.

    """

    abs_diff_means = np.abs(diff_means(array_one, array_two))

    return abs_diff_means


"""

# Histogram-weighted Networks (hiwenet)

"""

__all__ = ['extract', 'pairwise_dist', 'run_cli', 'more_metrics', 'non_pairwise']

from sys import version_info

if version_info.major==2 and version_info.minor==7:
    import more_metrics
    import non_pairwise
    from pairwise_dist import extract, run_cli
elif version_info.major > 2:
    from hiwenet import more_metrics, non_pairwise
    from hiwenet.pairwise_dist import extract, run_cli
else:
    raise NotImplementedError('hiwenet supports only 2.7 or 3+. Upgrade to Python 3+ is recommended.')

from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""

Script to analyze the relationships among different histogram distances,
    at different dimensionalities with different number of bins.

"""
import os
import pickle
from os.path import join as pjoin, exists as pexists

import numpy as np
from joblib import Parallel, delayed
from matplotlib import pyplot as plt
from medpy.metric import histogram

metric_list = ['manhattan', 'minowski', 'euclidean', 'noelle_2', 'noelle_4', 'noelle_5']
metric_list_to_doublecheck = ['cosine_1']
unknown_property = ['histogram_intersection']
still_under_dev  = ['quadratic_forms']
similarity_funcs = ['correlate', 'cosine', 'cosine_2', 'cosine_alt', 'fidelity_based']
semi_metric_list = ['jensen_shannon', 'chi_square', 'chebyshev', 'chebyshev_neg',
                    'histogram_intersection_1', 'relative_deviation', 'relative_bin_deviation',
                    'noelle_1', 'noelle_3', 'correlate_1']
excluded = ['kullback_leibler', ]

# all_methods = metric_list + similarity_funcs + semi_metric_list
# sticking to fewer methods with
all_methods = np.array(['chebyshev', 'chi_square', 'correlate', 'cosine', 'euclidean',
                           'histogram_intersection', 'jensen_shannon', 'manhattan', 'minowski',  'relative_deviation'])
all_methods.sort()

default_feature_dim = 1000
default_num_bins = 20
default_stdev = 0.1

range_dim = np.logspace(1, 3, num=20, dtype=int) # range(10, 100000, 1000)
range_num_bins = np.array(list(range(5, 200, 20)))
num_trials = 25

num_methods = len(all_methods)
num_dimensions = len(range_dim)
num_num_bins = len(range_num_bins)

num_rows = 3  # 7
num_cols = 7  # 3
fig_size = [14, 18]
num_ticks = 10

range_separability = [-0.75, -0.5, -0.25, 0.0, 0.25, 0.5, 0.75]
center_one = 0.0

edge_range = (-1.0, 1.0)
common_bin_edges = lambda nb: np.linspace(edge_range[0], edge_range[1], nb)

num_sep = len(range_separability)

rng = np.random.RandomState(654321)

def get_distr(center=0.0, stdev=default_stdev, length=50):
    "Returns a PDF of a given length. "
    # distr = np.random.random(length)

    # sticking to normal distibution to easily control separability
    distr = rng.normal(center, stdev, size=[length, 1])

    return distr


def make_random_histogram(center=0.0, stdev=default_stdev, length=default_feature_dim, num_bins=default_num_bins):
    "Returns a sequence of histogram density values that sum to 1.0"

    hist, bin_edges = np.histogram(get_distr(center, stdev, length),
                                   range=edge_range, bins=num_bins, density=True)
    # to ensure they sum to 1.0
    hist = hist / sum(hist)

    if len(hist) < 2:
        raise ValueError('Invalid histogram')

    return hist, bin_edges


def dist_betn_rand_hist(method, center_one, center_two, feat_dim, num_bins):

    h1, _ = make_random_histogram(center_one, length=feat_dim, num_bins=num_bins)
    h2, _ = make_random_histogram(center_two, length=feat_dim, num_bins=num_bins)

    return method(h1, h2)


# -------------------

print('range of dimensions : {}'.format(range_dim))

out_dir = '/data1/strother_lab/praamana/dist_compare_hiwenet_lowerdim_finegrain'
if not pexists(out_dir):
    os.mkdir(out_dir)

out_dir_analysis = '/data1/strother_lab/praamana/dist_compare_hiwenet_lowerdim_finegrain/analysis'
if not pexists(out_dir_analysis):
    os.mkdir(out_dir_analysis)

print('Computation ...')
for separability in range_separability:
    print('separability : {}'.format(separability))

    center_two = center_one + separability

    expt_id = 'separability{}_stdev{}_distances_dim_num_bins'.format(separability, default_stdev)
    saved_path = pjoin(out_dir,'{}.pkl'.format(expt_id))
    try:
        if pexists(saved_path):
            print('reading from disk')
            with open(saved_path, 'rb') as df:
                distances = pickle.load(df)
        else:
            raise IOError(' no distaces saved ')
    except:
        print('recomputing.')
        distances = np.full([num_methods, num_dimensions, num_num_bins, num_trials ], np.nan)

        for mm, method_str in enumerate(all_methods):
            method = getattr(histogram, method_str)

            print('Analyzing {} ... dim : '.format(method_str))
            for dd, feat_dim in enumerate(range_dim):
                print(' {} '.format(feat_dim), end='')
                for nn, num_bins in enumerate(range_num_bins):
                    distances[mm, dd, nn, :] = Parallel(n_jobs=num_trials)(delayed(dist_betn_rand_hist)(method, center_one, center_two, feat_dim, num_bins) for tt in range(num_trials))

            print(' .. Done.')

        with open(saved_path, 'wb') as df:
            pickle.dump(distances, df)

    print('computation for separability {} done.'.format(separability))

distances = np.full([num_sep, num_methods, num_dimensions, num_num_bins, num_trials ], np.nan)
for ss, separability in enumerate(range_separability):
    expt_id = 'separability{}_stdev{}_distances_dim_num_bins'.format(separability, default_stdev)
    saved_path = pjoin(out_dir,'{}.pkl'.format(expt_id))
    with open(saved_path, 'rb') as df:
        distances[ss, :, :, :, :] = pickle.load(df)

# ---------
print('Generating figures now... ')
for ss, separability in enumerate(range_separability):

    fig, ax = plt.subplots(num_rows, num_cols, figsize=fig_size)

    expt_id = 'separability{}_stdev{}_distances_dim_num_bins'.format(separability, default_stdev)

    tmp_vec = distances[ss, :, :, :, :].flatten()
    clim_max = np.percentile(tmp_vec, 99.5)
    clim_min = np.percentile(tmp_vec, 0.05)
    for mm, method_str in enumerate(all_methods):
        img_data = np.squeeze(np.median(distances[ss, mm, :, :, :], axis=2))

        ax = plt.subplot(num_rows, num_cols, mm + 1)
        img = plt.imshow(img_data) # , vmin=clim_min, vmax=clim_max)

        plt.title('{}'.format(method_str))
        tick_loc_dim = range(1, num_dimensions, int(np.floor(num_dimensions/num_ticks)))
        tick_loc_bins = range(1, num_num_bins, int(np.floor(num_num_bins/num_ticks)))

        if mm in range(0, num_methods, num_cols):
            xtick_loc = tick_loc_dim
            xtick_str = range_dim[tick_loc_dim]
        else:
            xtick_loc = []
            xtick_str = []

        ax.set_yticks(xtick_loc)
        ax.set_yticklabels(xtick_str)

        if mm > num_methods-num_cols-1:
            ytick_loc = tick_loc_bins
            ytick_str = range_num_bins[tick_loc_bins]
        else:
            ytick_loc = []
            ytick_str = []

        ax.set_xticks(ytick_loc)
        ax.set_xticklabels(ytick_str, rotation='vertical')

        if mm == 0:
            plt.ylabel('dimensionality ')

    plt.xlabel('num bins')
    plt.suptitle('separability : {}'.format(separability))
    out_fig_path = pjoin(out_dir_analysis, 'fig_{}.pdf'.format(expt_id))
    fig.savefig(out_fig_path, dpi=300)
    plt.close()
    # plt.show(block=False)


# -----------------

num_rows2, num_cols2 = 2, 4
for mm, method_str in enumerate(all_methods):
    fig, ax = plt.subplots(num_rows2, num_cols2, figsize=fig_size)

    tmp_vec = distances[:, mm, :, :, :].flatten()
    clim_max = np.percentile(tmp_vec, 99.5)
    clim_min = np.percentile(tmp_vec, 0.05)

    for ss, separability in enumerate(range_separability):
        img_data = np.squeeze(np.median(distances[ss, mm, :, :, :], axis=2))
        ax = plt.subplot(num_rows2, num_cols2, ss + 1)
        img = plt.imshow(img_data, vmin=clim_min, vmax=clim_max)
        plt.title('sep = {}'.format(separability))

    plt.suptitle('method : {}'.format(method_str))
    out_fig_path = pjoin(out_dir_analysis, 'method_{}_diff_seprabilities.pdf'.format(method_str))
    fig.savefig(out_fig_path, dpi=300)
    plt.close()



"""
Module collecting all the special methods that fit in the category of inter-regional distance metrics, but not pair-wise.

"""

__all__ = ['relative_to_all', ]

import numpy as np
import networkx as nx
from sys import version_info
if version_info.major==2 and version_info.minor==7:
    from .utils import compute_histogram
elif version_info.major > 2:
    from hiwenet.utils import compute_histogram
else:
    raise NotImplementedError('hiwenet supports only 2.7 or 3+. Upgrade to Python 3+ is recommended.')


def relative_to_all(features, groups, bin_edges, weight_func,
                    use_orig_distr,
                    group_ids, num_groups,
                    return_networkx_graph, out_weights_path):
    """
    Computes the given function (aka weight or distance) between histogram from each of the groups to a "grand histogram" derived from all groups.

    Parameters
    ----------
    features : ndarray or str
        1d array of scalar values, either provided directly as a 1d numpy array,
        or as a path to a file containing these values

    groups : ndarray or str
        Membership array of same length as `features`, each value specifying which group that particular node belongs to.
        Input can be either provided directly as a 1d numpy array,or as a path to a file containing these values.

        For example, if you have cortical thickness values for 1000 vertices (`features` is ndarray of length 1000),
        belonging to 100 patches, the groups array (of length 1000) could  have numbers 1 to 100 (number of unique values)
        specifying which element belongs to which cortical patch.

        Grouping with numerical values (contiguous from 1 to num_patches) is strongly recommended for simplicity,
        but this could also be a list of strings of length p, in which case a tuple is returned,
        identifying which weight belongs to which pair of patches.

    bin_edges : list or ndarray
        Array of bin edges within which to compute the histogram in.

    weight_func : callable
        Function to compute the edge weight between groups/nodes.

    use_orig_distr : bool, optional
        When using a user-defined callable, this flag
        1) allows skipping of pre-processing (trimming outliers) and histogram construction,
        2) enables the application of arbitrary callable (user-defined) on the original distributions coming from the two groups/ROIs/nodes directly.

        Example: ``diff_in_medians = lambda x, y: abs(np.median(x)-np.median(y))``

        This option is valid only when weight_method is a valid callable,
            which must take two inputs (possibly of different lengths) and return a single scalar.

    group_ids : list
        List of unique group ids to construct the nodes from (must all be present in the `groups` argument)

    num_groups : int
        Number of unique groups in the `group_ids`

    return_networkx_graph : bool, optional
        Specifies the need for a networkx graph populated with weights computed. Default: False.

    out_weights_path : str, optional
        Where to save the extracted weight matrix. If networkx output is returned, it would be saved in GraphML format.
        Default: nothing saved unless instructed.

    Returns
    -------
    distance_vector : ndarray
        vector of distances between the grand histogram and the individual ROIs

    Raises
    ------
    ValueError
        If one or more of the arrays are empty.

    """

    # notice the use of all features without regard to group membership
    hist_whole = compute_histogram(features, bin_edges, use_orig_distr)

    # to identify the central node capturing distribution from all roi's
    whole_node = 'whole'

    if return_networkx_graph:
        graph = nx.Graph()
        graph.add_nodes_from(group_ids)
        graph.add_node(whole_node)
    else:
        edge_weights = np.full([num_groups, 1], np.nan)

    for src in range(num_groups):
        index_roi = groups == group_ids[src]
        hist_roi = compute_histogram(features[index_roi], bin_edges, use_orig_distr)
        edge_value = weight_func(hist_whole, hist_roi)
        if return_networkx_graph:
            graph.add_edge(group_ids[src], whole_node, weight=float(edge_value))
        else:
            edge_weights[src] = edge_value

    if return_networkx_graph:
        if out_weights_path is not None:
            graph.write_graphml(out_weights_path)
        return graph
    else:
        if out_weights_path is not None:
            np.savetxt(out_weights_path, edge_weights, delimiter=',', fmt='%.9f')
        return edge_weights






import sys
import os
import shlex
import numpy as np
import scipy.stats
import networkx as nx
import matplotlib.pyplot as plt
from os.path import join as pjoin, exists as pexists, abspath
from sys import version_info

if version_info.major==2 and version_info.minor==7:
    from .pairwise_dist import extract as hiwenet
    from .pairwise_dist import run_cli as CLI
    from .pairwise_dist import metric_list, semi_metric_list
    from .utils import HiwenetWarning
elif version_info.major > 2:
    from hiwenet import extract as hiwenet
    from hiwenet import run_cli as CLI
    from hiwenet.pairwise_dist import metric_list, semi_metric_list
    from hiwenet.utils import HiwenetWarning
else:
    raise NotImplementedError('hiwenet supports only 2.7.13 or 3+. Upgrate to Python 3+ is recommended.')

list_weight_methods = metric_list + semi_metric_list
sys.dont_write_bytecode = True

from pytest import raises, warns, set_trace

dimensionality = np.random.randint(500, 2500)
num_groups = np.random.randint(2, 100)

def make_features(dimensionality, num_groups):
    num_links = int(num_groups*(num_groups-1)/2.0)

    group_ids_init = np.arange(num_groups)
    random_indices_into_groups = np.random.randint(0, num_groups, [1, dimensionality])
    groups = group_ids_init[random_indices_into_groups].flatten()

    group_ids = np.unique(groups)
    num_groups = len(group_ids)

    features = 10*np.random.random(dimensionality)

    return features, groups, group_ids, num_groups

features, groups, group_ids, num_groups = make_features(dimensionality, num_groups)
num_links = np.int64(num_groups * (num_groups - 1) / 2.0)
# for ix1, gid1 in enumerate(group_ids):
#     for ix2, gid2 in enumerate(group_ids):
#         if ix1 != ix2:
#             dist = cdist(features[groups==gid1],features[groups==gid2], 'euclidean')

cur_dir = os.path.dirname(abspath(__file__))

# the following are mostly usage tests. Refer to test_medpy.py for scientific validity of histogram metrics

def test_directed_mat():
    ew = hiwenet(features, groups, asymmetric=True)
    assert len(ew) == num_groups
    assert ew.shape[0] == num_groups and ew.shape[1] == num_groups

    # ensure no self-loops
    diag_elements = ew.flatten()[::num_groups+1]
    assert np.allclose(diag_elements, np.nan, equal_nan=True)

def test_directed_nx():
    graph = hiwenet(features, groups, asymmetric=True, return_networkx_graph=True)
    assert graph.is_directed()
    assert graph.number_of_nodes() == num_groups
    assert graph.number_of_edges() == 2*num_links

def test_dimensions():
    ew = hiwenet(features, groups)
    assert len(ew) == num_groups
    assert ew.shape[0] == num_groups and ew.shape[1] == num_groups

def test_more_metrics():
    ew = hiwenet(features, groups, weight_method='diff_medians',
                 use_original_distribution=True)
    assert len(ew) == num_groups
    assert ew.shape[0] == num_groups and ew.shape[1] == num_groups

    ew_abs = hiwenet(features, groups, weight_method='diff_medians_abs',
                     use_original_distribution=True)
    assert np.allclose(np.abs(ew), ew_abs, equal_nan=True)

    with warns(HiwenetWarning):
        ew = hiwenet(features, groups, weight_method='diff_medians',
                     use_original_distribution=False)

    with warns(HiwenetWarning):
        ew = hiwenet(features, groups,
                     weight_method='manhattan',
                     use_original_distribution=True)

def test_too_few_groups():
    features, groups, group_ids, num_groups = make_features(100, 1)
    with raises(ValueError):
        ew = hiwenet(features, groups)

def test_too_few_values():
    features, groups, group_ids, num_groups = make_features(10, 500)
    with raises(ValueError):
        ew = hiwenet(features[:num_groups-1], groups)

def test_invalid_trim_perc():

    with raises(ValueError):
        ew = hiwenet(features, groups, trim_percentile= -1)

    with raises(ValueError):
        ew = hiwenet(features, groups, trim_percentile=101)

def test_invalid_edge_range():

    with raises(ValueError):
        ew = hiwenet(features, groups, edge_range= -1)

    with raises(ValueError):
        ew = hiwenet(features, groups, edge_range=[])

    with raises(ValueError):
        ew = hiwenet(features, groups, edge_range=[1, ])

    with raises(ValueError):
        ew = hiwenet(features, groups, edge_range=[1, 2, 3])

    with raises(ValueError):
        ew = hiwenet(features, groups, edge_range=(1, np.NaN))

    with raises(ValueError):
        ew = hiwenet(features, groups, edge_range=(2, 1))

def test_invalid_weight_method():

    with raises(NotImplementedError):
        ew = hiwenet(features, groups, weight_method= 'dkjz.jscn')

    with raises(NotImplementedError):
        ew = hiwenet(features, groups, weight_method= 'somerandomnamenoonewoulduse')

def test_trim_not_too_few_values():
    with raises(ValueError):
        ew = hiwenet( [0], [1], trim_outliers = False)

def test_trim_false_too_few_to_calc_range():
    with raises(ValueError):
        ew = hiwenet( [1], groups, trim_outliers = False)

def test_not_np_arrays():
    with raises(ValueError):
        ew = hiwenet(list(), groups, trim_percentile=101)

    with raises(ValueError):
        ew = hiwenet(features, list(), trim_percentile=101)

def test_invalid_nbins():
    with raises(ValueError):
        ew = hiwenet(features, groups, num_bins=np.NaN)

    with raises(ValueError):
        ew = hiwenet(features, groups, num_bins=np.Inf)

    with raises(ValueError):
        ew = hiwenet(features, groups, num_bins=2)

def test_return_nx_graph():
    nxG = hiwenet(features, groups, return_networkx_graph = True)
    assert isinstance(nxG, nx.Graph)
    assert nxG.number_of_nodes() == num_groups
    assert nxG.number_of_edges() == num_links

def test_extreme_skewed():
    # Not yet sure what to test for here!!
    ew = hiwenet(10+np.zeros(dimensionality), groups)


# CLI tests
def test_CLI_run():
    "function to hit the CLI lines."

    # first word is the script names (ignored)
    
    featrs_path = abspath(pjoin(cur_dir, '..', 'examples', 'features_1000.txt'))
    groups_path = abspath(pjoin(cur_dir, '..', 'examples', 'groups_1000.txt'))
    sys.argv = shlex.split('hiwenet -f {} -g {} -n 25'.format(featrs_path, groups_path))
    CLI()

def test_CLI_output_matches_API():
    " Ensuring results from the two interfaces matches within a tolerance"

    # turning groups into strings to correspond with CLI
    groups_str = np.array([str(grp) for grp in groups])

    featrs_path = abspath(pjoin(cur_dir, '..', 'examples', 'test_features.txt'))
    groups_path = abspath(pjoin(cur_dir, '..', 'examples', 'test_groups.txt'))
    result_path = abspath(pjoin(cur_dir, '..', 'examples', 'test_result.txt'))
    np.savetxt(featrs_path, features, fmt='%20.9f')
    np.savetxt(groups_path, groups,  fmt='%d')

    for weight_method in list_weight_methods:
        api_result = hiwenet(features, groups_str, weight_method=weight_method)

        sys.argv = shlex.split('hiwenet -f {} -g {} -o {} -w {}'.format(featrs_path, groups_path, result_path, weight_method))
        CLI()
        cli_result = np.genfromtxt(result_path, delimiter=',')

        if not bool(np.allclose(cli_result, api_result, rtol=1e-2, atol=1e-3, equal_nan=True)):
            raise ValueError('CLI results differ from API for {}'.format(weight_method))


def test_CLI_nonexisting_paths():
    "invalid paths"

    featrs_path = abspath(pjoin(cur_dir, '..', 'examples', 'features_1000.txt'))
    groups_path = 'NONEXISTING_groups_1000.txt'
    sys.argv = shlex.split('hiwenet -f {} -g {} -n 25'.format(featrs_path, groups_path))
    with raises(IOError):
        CLI()

    featrs_path = 'NONEXISTING_features_1000.txt'
    groups_path = abspath(pjoin(cur_dir, '..', 'examples', 'groups_1000.txt'))
    sys.argv = shlex.split('hiwenet -f {} -g {} -n 25'.format(featrs_path, groups_path))
    with raises(IOError):
        CLI()


def test_CLI_invalid_args():
    "invalid paths"

    featrs_path = 'NONEXISTING_features_1000.txt'
    # arg aaa or invalid_arg_name doesnt exist
    sys.argv = shlex.split('hiwenet --aaa {0} -f {0} -g {0}'.format(featrs_path))
    with raises(SystemExit):
        CLI()

    sys.argv = shlex.split('hiwenet --invalid_arg_name {0} -f {0} -g {0}'.format(featrs_path))
    with raises(SystemExit):
        CLI()


def test_CLI_too_few_args():
    "testing too few args"

    sys.argv = ['hiwenet ']
    with raises(SystemExit):
        CLI()

    sys.argv = ['hiwenet -f check']
    with raises(SystemExit):
        CLI()

    sys.argv = ['hiwenet -g check']
    with raises(SystemExit):
        CLI()


def test_input_callable():

    diff_skew = lambda x, y: abs(scipy.stats.skew(x)-scipy.stats.skew(y))
    ew = hiwenet(features, groups, weight_method=diff_skew)

    assert len(ew) == num_groups
    assert ew.shape[0] == num_groups and ew.shape[1] == num_groups


def test_input_callable_on_orig_data():

    diff_medians = lambda x, y: abs(np.median(x)-np.median(y))
    ew = hiwenet(features, groups, weight_method=diff_medians,
                 use_original_distribution=True)

    assert len(ew) == num_groups
    assert ew.shape[0] == num_groups and ew.shape[1] == num_groups

    # checking for invalid use : histogram metrics can not be applied on orig data (without building histograms)
    with warns(HiwenetWarning):
        ew = hiwenet(features, groups,
                     weight_method='manhattan',
                     use_original_distribution=True)

def test_relative_to_all():
    ""

    for weight_method in list_weight_methods:
        ew = hiwenet(features, groups,
                     weight_method=weight_method,
                     relative_to_all=True)

        # it must be a vector
        assert ew.size == num_groups
        assert ew.shape[0] == num_groups
        assert ew.shape[1] == 1

        nxg = hiwenet(features, groups,
                      weight_method=weight_method,
                      relative_to_all=True,
                      return_networkx_graph=True)

        # additional node is the grand node (from all roi's)
        assert nxg.number_of_nodes() == num_groups+1
        assert nxg.number_of_edges() == num_groups

        weights_from_nxg = np.array([ nxg[src]['whole']['weight'] for src in group_ids ])
        assert np.allclose(ew.flatten(), weights_from_nxg)



# test_directed_nx()
# test_directed_mat()
# test_CLI_output_matches_API()
# test_input_callable()

# test_more_metrics()

test_relative_to_all()


#!/usr/bin/env python

from setuptools import setup, find_packages
import versioneer

setup(name='hiwenet',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Histogram-weighted Networks for Feature Extraction and Advance Analysis in Neuroscience',
      long_description='Histogram-weighted Networks for Feature Extraction and Advance Analysis in Neuroscience; hiwenet',
      author='Pradeep Reddy Raamana',
      author_email='raamana@gmail.com',
      url='https://github.com/raamana/hiwenet',
      packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]), # ['neuropredict'],
      install_requires=['numpy', 'pyradigm', 'nibabel', 'networkx', 'medpy'],
      classifiers=[
              'Intended Audience :: Science/Research',
              'Programming Language :: Python',
              'Topic :: Software Development',
              'Topic :: Scientific/Engineering',
              'Operating System :: Microsoft :: Windows',
              'Operating System :: POSIX',
              'Operating System :: Unix',
              'Operating System :: MacOS',
              'Programming Language :: Python :: 3.5',
              'Programming Language :: Python :: 3.6',              
              'Programming Language :: Python :: 2.7',
          ],
      entry_points={
          "console_scripts": [
              "hiwenet=hiwenet.__main__:main",
          ]
      }

     )



from sys import version_info

if version_info.major==2 and version_info.minor==7:
    from pairwise_dist import run_cli
elif version_info.major > 2:
    from hiwenet.pairwise_dist import run_cli
else:
    raise NotImplementedError('hiwenet supports only 2.7.13 or 3+. Upgrade to Python 3+ is recommended.')

def main():
    "Entry point."

    run_cli()

if __name__ == '__main__':
    main()



__all__ = ['extract', 'run_cli']

import argparse
import os
import sys
import collections
import traceback
import warnings
import logging
import networkx as nx
import numpy as np
from os.path import join as pjoin, exists as pexists
from sys import version_info

if version_info.major==2 and version_info.minor==7:
    import more_metrics
    import non_pairwise
    from .utils import compute_histogram, HiwenetWarning
elif version_info.major > 2:
    from hiwenet import more_metrics, non_pairwise
    from hiwenet.utils import compute_histogram, HiwenetWarning
else:
    raise NotImplementedError('hiwenet supports only 2.7 or 3+. Upgrade to Python 3+ is recommended.')

list_medpy_histogram_metrics = np.array([
    'chebyshev', 'chebyshev_neg', 'chi_square',
    'correlate', 'correlate_1',
    'cosine', 'cosine_1', 'cosine_2', 'cosine_alt',
    'euclidean', 'fidelity_based',
    'histogram_intersection', 'histogram_intersection_1',
    'jensen_shannon', 'kullback_leibler', 'manhattan', 'minowski',
    'noelle_1', 'noelle_2', 'noelle_3', 'noelle_4', 'noelle_5',
    'relative_bin_deviation', 'relative_deviation'])

metric_list = [
    'manhattan', 'minowski', 'euclidean',
    'noelle_2', 'noelle_4', 'noelle_5']

unknown_prop_list = ['histogram_intersection']
still_under_dev = ['quadratic_forms']
similarity_func = ['kullback_leibler', 'cosine_1', 'correlate', 'cosine', 'cosine_2', 'cosine_alt', 'fidelity_based']

semi_metric_list = [
    'jensen_shannon', 'chi_square',
    'chebyshev', 'chebyshev_neg',
    'histogram_intersection_1',
    'relative_deviation', 'relative_bin_deviation',
    'noelle_1', 'noelle_3',
    'correlate_1']

metrics_on_original_features = ['diff_medians', 'diff_medians_abs',
                                'diff_means',   'diff_means_abs' ]

symmetric_metrics_on_original_features = ['diff_medians_abs', 'diff_means_abs']

minimum_num_bins = 5

default_weight_method = 'manhattan'
default_num_bins = 25
default_edge_range = None
default_trim_percentile = 5
default_trim_behaviour = True
default_return_networkx_graph = False
default_out_weights_path = None

def compute_bin_edges(features, num_bins, edge_range, trim_outliers, trim_percentile, use_orig_distr=False):
    "Compute the edges for the histogram bins to keep it the same for all nodes."

    if use_orig_distr:
        print('Using original distribution (without histogram) to compute edge weights!')
        edges=None
        return edges

    if edge_range is None:
        if trim_outliers:
            # percentiles_to_keep = [ trim_percentile, 1.0-trim_percentile] # [0.05, 0.95]
            edges_of_edges = np.array([np.percentile(features, trim_percentile),
                                       np.percentile(features, 100 - trim_percentile)])
        else:
            edges_of_edges = np.array([np.min(features), np.max(features)])
    else:
        edges_of_edges = edge_range

    # Edges computed using data from all nodes, in order to establish correspondence
    edges = np.linspace(edges_of_edges[0], edges_of_edges[1], num=num_bins, endpoint=True)

    return edges


def extract(features, groups,
            weight_method=default_weight_method,
            num_bins=default_num_bins,
            edge_range=default_edge_range,
            trim_outliers=default_trim_behaviour,
            trim_percentile=default_trim_percentile,
            use_original_distribution=False,
            relative_to_all=False,
            asymmetric=False,
            return_networkx_graph=default_return_networkx_graph,
            out_weights_path=default_out_weights_path):
    """
    Extracts the histogram-distance weighted adjacency matrix.

    Parameters
    ----------
    features : ndarray or str
        1d array of scalar values, either provided directly as a 1d numpy array,
        or as a path to a file containing these values

    groups : ndarray or str
        Membership array of same length as `features`, each value specifying which group that particular node belongs to.
        Input can be either provided directly as a 1d numpy array,or as a path to a file containing these values.

        For example, if you have cortical thickness values for 1000 vertices (`features` is ndarray of length 1000),
        belonging to 100 patches, the groups array (of length 1000) could  have numbers 1 to 100 (number of unique values)
        specifying which element belongs to which cortical patch.

        Grouping with numerical values (contiguous from 1 to num_patches) is strongly recommended for simplicity,
        but this could also be a list of strings of length p, in which case a tuple is returned,
        identifying which weight belongs to which pair of patches.

    weight_method : string or callable, optional
        Type of distance (or metric) to compute between the pair of histograms.
        It can either be a string identifying one of the weights implemented below, or a valid callable.

        If a string, it must be one of the following methods:

        - 'chebyshev'
        - 'chebyshev_neg'
        - 'chi_square'
        - 'correlate'
        - 'correlate_1'
        - 'cosine'
        - 'cosine_1'
        - 'cosine_2'
        - 'cosine_alt'
        - 'euclidean'
        - 'fidelity_based'
        - 'histogram_intersection'
        - 'histogram_intersection_1'
        - 'jensen_shannon'
        - 'kullback_leibler'
        - 'manhattan'
        - 'minowski'
        - 'noelle_1'
        - 'noelle_2'
        - 'noelle_3'
        - 'noelle_4'
        - 'noelle_5'
        - 'relative_bin_deviation'
        - 'relative_deviation'

        Note only the following are *metrics*:

        - 'manhattan'
        - 'minowski'
        - 'euclidean'
        - 'noelle_2'
        - 'noelle_4'
        - 'noelle_5'

        The following are *semi- or quasi-metrics*:

        - 'kullback_leibler'
        - 'jensen_shannon'
        - 'chi_square'
        - 'chebyshev'
        - 'cosine_1'
        - 'chebyshev_neg'
        - 'correlate_1'
        - 'histogram_intersection_1'
        - 'relative_deviation'
        - 'relative_bin_deviation'
        - 'noelle_1'
        - 'noelle_3'

        The following are  classified to be similarity functions:

        - 'histogram_intersection'
        - 'correlate'
        - 'cosine'
        - 'cosine_2'
        - 'cosine_alt'
        - 'fidelity_based'

        *Default* choice: 'minowski'.

        The method can also be one of the following identifying metrics that operate on the original data directly -
         e.g. difference in the medians coming from the distributions of the pair of ROIs.

         - 'diff_medians'
         - 'diff_means'
         - 'diff_medians_abs'
         - 'diff_means_abs'

         Please note this can lead to adjacency matrices that may not be symmetric
            e.g. difference metric on two scalars is not symmetric).
            In this case, be sure to use the flag: allow_non_symmetric=True

        If weight_method is a callable, it must two accept two arrays as input and return one scalar as output.
            Example: ``diff_in_skew = lambda x, y: abs(scipy.stats.skew(x)-scipy.stats.skew(y))``
            NOTE: this method will be applied to histograms (not the original distribution of features from group/ROI).
            In order to apply this callable directly on the original distribution (without trimming and histogram binning),
            use ``use_original_distribution=True``.

    num_bins : scalar, optional
        Number of bins to use when computing histogram within each patch/group.

        Note:

        1) Please ensure same number of bins are used across different subjects
        2) histogram shape can vary widely with number of bins (esp with fewer bins in the range of 3-20), and hence the features extracted based on them vary also.
        3) It is recommended to study the impact of this parameter on the final results of the experiment.

        This could also be optimized within an inner cross-validation loop if desired.

    edge_range : tuple or None
        The range of edges within which to bin the given values.
        This can be helpful to ensure correspondence across multiple invocations of hiwenet (for different subjects),
        in terms of range across all bins as well as individual bin edges.
        Default is to automatically compute from the given values.

        Accepted format:

            - tuple of finite values: (range_min, range_max)
            - None, triggering automatic calculation (default)

        Notes : when controlling the ``edge_range``, it is not possible trim the tails (e.g. using the parameters
        ``trim_outliers`` and ``trim_percentile``) for the current set of features using its own range.

    trim_outliers : bool, optional
        Whether to trim a small percentile of outliers at the edges of feature range,
        when features are expected to contain extreme outliers (like 0 or eps or Inf).
        This is important to avoid numerical problems and also to stabilize the weight estimates.

    trim_percentile : float
        Small value specifying the percentile of outliers to trim.
        Default: 5 (5%). Must be in open interval (0, 100).

    use_original_distribution : bool, optional
        When using a user-defined callable, this flag
        1) allows skipping of pre-processing (trimming outliers) and histogram construction,
        2) enables the application of arbitrary callable (user-defined) on the original distributions coming from the two groups/ROIs/nodes directly.

        Example: ``diff_in_medians = lambda x, y: abs(np.median(x)-np.median(y))``

        This option is valid only when weight_method is a valid callable,
            which must take two inputs (possibly of different lengths) and return a single scalar.

    relative_to_all : bool
        Flag to instruct the computation of a grand histogram (distribution pooled from values in all ROIs),
        and compute distances (based on distance specified by ``weight_method``) by from each ROI to the grand mean.
        This would result in only N distances for N ROIs, instead of the usual N*(N-1) pair-wise distances.

    asymmetric : bool
        Flag to identify resulting adjacency matrix is expected to be non-symmetric.
        Note: this results in twice the computation time!
        Default: False , for histogram metrics implemented here are symmetric.

    return_networkx_graph : bool, optional
        Specifies the need for a networkx graph populated with weights computed. Default: False.

    out_weights_path : str, optional
        Where to save the extracted weight matrix. If networkx output is returned, it would be saved in GraphML format.
        Default: nothing saved unless instructed.

    Returns
    -------
    edge_weights : ndarray
        numpy 2d array of pair-wise edge-weights (of size: num_groups x num_groups),
        wherein num_groups is determined by the total number of unique values in `groups`.

        **Note**:

        - Only the upper triangular matrix is filled as the distance between node i and j would be the same as j and i.
        - The edge weights from the upper triangular matrix can easily be obtained by

        .. code-block:: python

            weights_array = edge_weights[ np.triu_indices_from(edge_weights, 1) ]

    """

    # parameter check
    features, groups, num_bins, edge_range, group_ids, num_groups, num_links = check_params(
            features, groups, num_bins, edge_range, trim_outliers, trim_percentile)

    weight_func, use_orig_distr, non_symmetric = check_weight_method(weight_method,
                                                                     use_original_distribution, asymmetric)

    # using the same bin edges for all nodes/groups to ensure correspondence
    # NOTE: common bin edges is important for the disances to be any meaningful
    edges = compute_bin_edges(features, num_bins, edge_range,
                              trim_outliers, trim_percentile, use_orig_distr)

    # handling special
    if relative_to_all:
        result = non_pairwise.relative_to_all(features, groups, edges, weight_func,
                                              use_orig_distr, group_ids, num_groups,
                                              return_networkx_graph, out_weights_path)
    else:
        result = pairwise_extract(features, groups, edges, weight_func, use_orig_distr,
                                  group_ids, num_groups, num_links,
                                  non_symmetric, return_networkx_graph, out_weights_path)

    # this can be a networkx graph or numpy array depending on request
    return result


def pairwise_extract(features, groups, edges, weight_func, use_orig_distr,
                     group_ids, num_groups, num_links,
                     non_symmetric, return_networkx_graph, out_weights_path):
    """
    Core function to compute the pair-wise histogram distance between all ROIs.

    Parameters
    ----------
    features
    groups
    edges
    weight_func
    use_orig_distr
    group_ids
    num_groups
    num_links
    non_symmetric
    return_networkx_graph
    out_weights_path

    Returns
    -------
    result : object
        A networkx graph or numpy array depending on request

    """

    # the following will execute only when the pair-wise computation is requested.
    if return_networkx_graph:
        graph = nx.DiGraph() if non_symmetric else nx.Graph()
        graph.add_nodes_from(group_ids)
    else:
        edge_weights = np.full([num_groups, num_groups], np.nan)

    exceptions_list = list()
    for src in range(num_groups):
        # primitive progress indicator
        if np.mod(src + 1, 5) == 0.0:
            sys.stdout.write('.')

        index1 = groups == group_ids[src]
        hist_one = compute_histogram(features[index1], edges, use_orig_distr)

        if non_symmetric:
            target_list = range(num_groups)
        else:
            # when symmetric, only upper tri matrix is computed/filled
            target_list = range(src + 1, num_groups, 1)

        for dest in target_list:
            # skipping edge between self
            if src == dest:
                continue

            index2 = groups == group_ids[dest]
            hist_two = compute_histogram(features[index2], edges, use_orig_distr)

            try:
                edge_value = weight_func(hist_one, hist_two)
                if return_networkx_graph:
                    graph.add_edge(group_ids[src], group_ids[dest], weight=float(edge_value))
                else:
                    edge_weights[src, dest] = edge_value
            except (RuntimeError, RuntimeWarning) as runexc:
                # placeholder to ignore some runtime errors (such as medpy's logger issue)
                print(runexc)
            except BaseException as exc:
                # numerical instabilities can cause trouble for histogram distance calculations
                traceback.print_exc()
                exceptions_list.append(str(exc))
                logging.warning('Unable to compute edge weight between '
                                ' {} and {}. Skipping it.'.format(group_ids[src], group_ids[dest]))

    error_thresh = 0.05
    if len(exceptions_list) >= error_thresh * num_links:
        print('All exceptions encountered so far:\n {}'.format('\n'.join(exceptions_list)))
        raise ValueError('Weights for atleast {:.2f}% of edges could not be computed.'.format(error_thresh * 100))

    sys.stdout.write('\n')

    if return_networkx_graph:
        if out_weights_path is not None:
            graph.write_graphml(out_weights_path)
        return graph
    else:
        if out_weights_path is not None:
            np.savetxt(out_weights_path, edge_weights, delimiter=',', fmt='%.9f')
        return edge_weights


def identify_groups(groups):
    """
    To compute number of unique elements in a given membership specification.

    Parameters
    ----------
    groups : numpy 1d array of length p, each value specifying which group that particular node belongs to.
        For examlpe, if you have a cortical thickness values for 1000 vertices belonging to 100 patches,
        this array could  have numbers 1 to 100 specifying which vertex belongs to which cortical patch.
        Although grouping with numerical values (contiguous from 1 to num_patches) is strongly recommended for simplicity,
        this could also be a list of strings of length p.

    Returns
    -------
    group_ids : numpy array of values identifying the unique groups specified
    num_groups : scalar value denoting the number of unique groups specified

    """

    group_ids = np.unique(groups)
    num_groups = len(group_ids)

    if num_groups < 2:
        raise ValueError('There must be atleast two nodes or groups in data, for pair-wise edge-weight calculations.')

    return group_ids, num_groups


def check_param_ranges(num_bins, num_groups, num_values, trim_outliers, trim_percentile):
    """Ensuring the parameters are in valid ranges."""

    if num_bins < minimum_num_bins:
        raise ValueError('Too few bins! The number of bins must be >= 5')

    if num_values < num_groups:
        raise ValueError('Insufficient number of values in features (< number of nodes), or invalid membership!')

    if trim_outliers:
        if trim_percentile < 0 or trim_percentile >= 100:
            raise ValueError('percentile of tail values to trim must be in the semi-open interval [0,1).')
    elif num_values < 2:
        raise ValueError('too few features to compute minimum and maximum')

    return


def type_cast_params(num_bins, edge_range_spec, features, groups):
    """Casting inputs to required types."""

    if isinstance(num_bins, str):
        # possible when called from CLI
        num_bins = np.float(num_bins)

    # rounding it to ensure it is int
    num_bins = np.rint(num_bins)

    if np.isnan(num_bins) or np.isinf(num_bins):
        raise ValueError('Invalid value for number of bins! Choose a natural number >= {}'.format(minimum_num_bins))

    if edge_range_spec is None:
        edge_range = edge_range_spec
    elif isinstance(edge_range_spec, collections.Sequence):
        if len(edge_range_spec) != 2:
            raise ValueError('edge_range must be a tuple of two values: (min, max)')
        if edge_range_spec[0] >= edge_range_spec[1]:
            raise ValueError('edge_range : min {} is not less than the max {} !'.format(edge_range_spec[0], edge_range_spec[1]))
        if not np.all(np.isfinite(edge_range_spec)):
            raise ValueError('Infinite or NaN values in edge range : {}'.format(edge_range_spec))

        # converting it to tuple to make it immutable
        edge_range = tuple(edge_range_spec)
    else:
        raise ValueError('Invalid edge range! Must be a tuple of two values (min, max)')

    if not isinstance(features, np.ndarray):
        features = np.array(features)

    if not isinstance(groups, np.ndarray):
        groups = np.array(groups)

    return num_bins, edge_range, features, groups


def make_random_histogram(length=100, num_bins=10):
    "Returns a sequence of histogram density values that sum to 1.0"

    hist, bin_edges = np.histogram(np.random.random(length),
                                   bins=num_bins, density=True)

    # to ensure they sum to 1.0
    hist = hist / sum(hist)

    if len(hist) < 2:
        raise ValueError('Invalid histogram')

    return hist


def check_weight_method(weight_method_spec,
                        use_orig_distr=False,
                        allow_non_symmetric=False):
    "Check if weight_method is recognized and implemented, or ensure it is callable."

    if not isinstance(use_orig_distr, bool):
        raise TypeError('use_original_distribution flag must be boolean!')

    if not isinstance(allow_non_symmetric, bool):
        raise TypeError('allow_non_symmetric flag must be boolean')

    if isinstance(weight_method_spec, str):
        weight_method_spec = weight_method_spec.lower()

        if weight_method_spec in list_medpy_histogram_metrics:
            from medpy.metric import histogram as medpy_hist_metrics
            weight_func = getattr(medpy_hist_metrics, weight_method_spec)
            if use_orig_distr:
                warnings.warn('use_original_distribution must be False when using builtin histogram metrics, '
                                 'which expect histograms as input - setting it to False.', HiwenetWarning)
                use_orig_distr = False

        elif weight_method_spec in metrics_on_original_features:
            weight_func = getattr(more_metrics, weight_method_spec)
            if not use_orig_distr:
                warnings.warn('use_original_distribution must be True when using builtin non-histogram metrics, '
                              'which expect original feature values in ROI/node as input '
                              '- setting it to True.', HiwenetWarning)
                use_orig_distr = True

            if weight_method_spec in symmetric_metrics_on_original_features:
                print('Chosen metric is symmetric. Ignoring asymmetric=False flag.')
                allow_non_symmetric=False

        else:
            raise NotImplementedError('Chosen histogram distance/metric not implemented or invalid.')

    elif callable(weight_method_spec):
        # ensure 1) takes two ndarrays
        try:
            dummy_weight = weight_method_spec(make_random_histogram(), make_random_histogram())
        except:
            raise TypeError('Error applying given callable on two input arrays.\n'
                            '{} must accept two arrays and return a single scalar value!')
        else:
            # and 2) returns only one number
            if not np.isscalar(dummy_weight):
                raise TypeError('Given callable does not return a single scalar as output.')

        weight_func = weight_method_spec
    else:
        raise ValueError('Supplied method to compute edge weight is not recognized:\n'
                         'must be a string identifying one of the implemented methods\n{}'
                         '\n or a valid callable that accepts that two arrays '
                         'and returns 1 scalar.'.format(list_medpy_histogram_metrics))

    return weight_func, use_orig_distr, allow_non_symmetric


def check_params(features_spec, groups_spec, num_bins, edge_range_spec, trim_outliers, trim_percentile):
    """Necessary check on values, ranges, and types."""

    if isinstance(features_spec, str) and isinstance(groups_spec, str):
        features, groups = read_features_and_groups(features_spec, groups_spec)
    else:
        features, groups = features_spec, groups_spec

    num_bins, edge_range, features, groups = type_cast_params(num_bins, edge_range_spec, features, groups)
    num_values = len(features)

    # memberships
    group_ids, num_groups = identify_groups(groups)
    num_links = np.int64(num_groups * (num_groups - 1) / 2.0)

    check_param_ranges(num_bins, num_groups, num_values, trim_outliers, trim_percentile)

    return features, groups, num_bins, edge_range, group_ids, num_groups, num_links


def run_cli():
    "Command line interface to hiwenet."

    features_path, groups_path, weight_method, num_bins, edge_range, \
    trim_outliers, trim_percentile, return_networkx_graph, out_weights_path = parse_args()

    # TODO add the possibility to process multiple combinations of parameters: diff subjects, diff metrics
    # for features_path to be a file containing multiple subjects (one/line)
    # -w could take multiple values kldiv,histint,
    # each line: input_features_path,out_weights_path

    features, groups = read_features_and_groups(features_path, groups_path)

    extract(features, groups, weight_method=weight_method, num_bins=num_bins,
            edge_range=edge_range, trim_outliers=trim_outliers, trim_percentile=trim_percentile,
            return_networkx_graph=return_networkx_graph, out_weights_path=out_weights_path)


def read_features_and_groups(features_path, groups_path):
    "Reader for data and groups"

    try:
        if not pexists(features_path):
            raise ValueError('non-existent features file')

        if not pexists(groups_path):
            raise ValueError('non-existent groups file')

        if isinstance(features_path, str):
            features = np.genfromtxt(features_path, dtype=float)
        else:
            raise ValueError('features input must be a file path ')

        if isinstance(groups_path, str):
            groups = np.genfromtxt(groups_path, dtype=str)
        else:
            raise ValueError('groups input must be a file path ')

    except:
        raise IOError('error reading the specified features and/or groups.')

    if len(features) != len(groups):
        raise ValueError("lengths of features and groups do not match!")

    return features, groups


def get_parser():
    "Specifies the arguments and defaults, and returns the parser."

    parser = argparse.ArgumentParser(prog="hiwenet")

    parser.add_argument("-f", "--in_features_path", action="store", dest="in_features_path",
                        required=True,
                        help="Abs. path to file containing features for a given subject")

    parser.add_argument("-g", "--groups_path", action="store", dest="groups_path",
                        required=True,
                        help="path to a file containing element-wise membership into groups/nodes/patches.")

    parser.add_argument("-w", "--weight_method", action="store", dest="weight_method",
                        default=default_weight_method, required=False,
                        help="Method used to estimate the weight between the pair of nodes. Default : {}".format(
                            default_weight_method))

    parser.add_argument("-o", "--out_weights_path", action="store", dest="out_weights_path",
                        default=default_out_weights_path, required=False,
                        help="Where to save the extracted weight matrix. If networkx output is returned, it would be saved in GraphML format. Default: nothing saved.")

    parser.add_argument("-n", "--num_bins", action="store", dest="num_bins",
                        default=default_num_bins, required=False,
                        help="Number of bins used to construct the histogram. Default : {}".format(default_num_bins))

    parser.add_argument("-r", "--edge_range", action="store", dest="edge_range",
                        default=default_edge_range, required=False,
                        nargs = 2,
                        help="The range of edges (two finite values) within which to bin the given values e.g. --edge_range 1 6 "
                             "This can be helpful to ensure correspondence across multiple invocations of hiwenet (for different subjects),"
                             " in terms of range across all bins as well as individual bin edges. "
                             "Default : {}, to automatically compute from the given values.".format(default_edge_range))

    parser.add_argument("-t", "--trim_outliers", action="store", dest="trim_outliers",
                        default=default_trim_behaviour, required=False,
                        help="Boolean flag indicating whether to trim the extreme/outlying values. Default True.")

    parser.add_argument("-p", "--trim_percentile", action="store", dest="trim_percentile",
                        default=default_trim_percentile, required=False,
                        help="Small value specifying the percentile of outliers to trim. "
                             "Default: {0}%% , must be in open interval (0, 100).".format(default_trim_percentile))

    parser.add_argument("-x", "--return_networkx_graph", action="store", dest="return_networkx_graph",
                        default=default_return_networkx_graph, required=False,
                        help="Boolean flag indicating whether to return a networkx graph populated with weights computed. Default: False")

    return parser


def parse_args():
    """Parser/validator for the cmd line args."""

    parser = get_parser()

    if len(sys.argv) < 2:
        parser.print_help()
        warnings.warn('Too few arguments!', UserWarning)
        parser.exit(1)

    # parsing
    try:
        params = parser.parse_args()
    except Exception as exc:
        print(exc)
        raise ValueError('Unable to parse command-line arguments.')

    in_features_path = os.path.abspath(params.in_features_path)
    if not os.path.exists(in_features_path):
        raise IOError("Given features file doesn't exist.")

    groups_path = os.path.abspath(params.groups_path)
    if not os.path.exists(groups_path):
        raise IOError("Given groups file doesn't exist.")

    return in_features_path, groups_path, params.weight_method, params.num_bins, params.edge_range, \
           params.trim_outliers, params.trim_percentile, params.return_networkx_graph, params.out_weights_path


if __name__ == '__main__':
    run_cli()
